#include "general/config.h"

#include <errno.h>
#include <string.h>
#include <time.h>

#include <string>
#include <iostream>
#include <sstream>

#include "general/DeadLockDetector.hh"
#include "general/IOLock.hh"
#include "general/mutexlock.hh"
#include "general/ReadWriteLock.hh"

using namespace General;

#define VERBOSE_RETRY_LOGGING 1

#if DEAD_LOCK_DETECTOR_ENABLED
using General::DeadLockDetector;
#endif /* DEAD_LOCK_DETECTOR_ENABLED */

namespace
{
#if ! HAVE_PTHREAD_RW_LOCK
  inline int
  pthread_rwlock_destroy( General::ReadWriteLock::lock_type* Lock )
  {
    return pthread_mutex_destroy( Lock );
  }

  inline int
  pthread_rwlock_init(  General::ReadWriteLock::lock_type* Lock,
			void* Attr )
  {
    return pthread_mutex_init( Lock,
			       static_cast< pthread_mutexattr_t* >( Attr ) );
  }

  inline int
  pthread_rwlock_unlock( General::ReadWriteLock::lock_type* Lock )
  {
    return pthread_mutex_unlock( Lock );
  }

  inline int
  pthread_rwlock_tryrdlock( General::ReadWriteLock::lock_type* Lock )
  {
    return pthread_mutex_trylock( Lock );
  }

  inline int
  pthread_rwlock_trywrlock( General::ReadWriteLock::lock_type* Lock )
  {
    return pthread_mutex_trylock( Lock );
  }

  inline int
  pthread_rwlock_rdlock( General::ReadWriteLock::lock_type* Lock )
  {
    return pthread_mutex_lock( Lock );
  }

  inline int
  pthread_rwlock_wrlock( General::ReadWriteLock::lock_type* Lock )
  {
    return pthread_mutex_trylock( Lock );
  }

  inline int
  pthread_rwlock_timedrdlock( General::ReadWriteLock::lock_type* Lock,
			      struct timespec*	Timeout )
  {
    return pthread_mutex_timedlock( Lock, Timeout );
  }

  inline int
  pthread_rwlock_timedwrlock( General::ReadWriteLock::lock_type* Lock,
			      struct timespec*	Timeout )
  {
    return pthread_mutex_timedlock( Lock, Timeout );
  }

#undef   PTHREAD_RWLOCK_INITIALIZER
#define  PTHREAD_RWLOCK_INITIALIZER PTHREAD_MUTEX_INITIALIZER
#endif /* ! HAVE_PTHREAD_RW_LOCK */

  struct timespec LockAcquisitionInterval = {
    0,
    250000000,
  };

  std::string
  log ( const int Error,
	const std::string& Message,
	const std::string& Level = "ERROR" )
  {
    std::ostringstream	msg;

    msg << Level << ": ReadWriteLock: " << Message
	<< " Error Code: " << Error
	<< " system error: " << strerror( Error );

    MutexLock::lock_type*	io_baton( IOLock::GetKey( std::cerr ) );

    MutexLock::Lock( io_baton );
    std::cerr << msg.str( ) << std::endl;
    MutexLock::UnLock( io_baton );
    return msg.str( );
  }

  inline void
  normalize( timespec& Time )
  {
    static const long	ns = 1000000000;
    if ( Time.tv_nsec > ns )
    {
      Time.tv_sec += Time.tv_sec / ns;	// Add second component
      Time.tv_nsec = Time.tv_nsec % ns;	// truncate seconds
    }
  }

  inline bool
  operator<( const struct timespec& LHS,
	     const struct timespec& RHS )
  {
    if ( LHS.tv_sec < RHS.tv_sec )
    {
      return true;
    }
    else if ( LHS.tv_sec > RHS.tv_sec )
    {
      return false;
    }
    else if ( LHS.tv_sec < RHS.tv_sec )
    {
      return true;
    }
    else
    {
      return false;
    }
  }

  inline struct timespec
  operator+=( struct timespec& LHS,
	      const struct timespec& RHS )
  {
    // Add two values together
    LHS.tv_sec += RHS.tv_sec;
    LHS.tv_nsec += RHS.tv_nsec;
    normalize( LHS );
    return LHS;
  }

  inline int
  get_lock( General::ReadWriteLock::lock_type* Lock,
	    General::ReadWriteLock::mode_type Mode,
	    bool TryLock )
  {
    int retval = 0;

    switch( Mode )
    {
    case General::ReadWriteLock::READ:
      retval = ( TryLock )
	? pthread_rwlock_tryrdlock( Lock )
	: pthread_rwlock_rdlock( Lock );
	break;
    case General::ReadWriteLock::WRITE:
      retval = ( TryLock )
	? pthread_rwlock_trywrlock( Lock )
	: pthread_rwlock_wrlock( Lock );
	break;
    }
    return retval;
  }

  inline int
  get_lock( General::ReadWriteLock::lock_type* Lock,
	    General::ReadWriteLock::mode_type Mode,
	    const struct timespec& RelTimeout,
	    const struct timespec& AbsStop )
  {
    int retval = 0;
    struct timespec	timeout;

    clock_gettime( CLOCK_REALTIME, &timeout );

    if ( AbsStop < timeout )
    {
      throw std::runtime_error( "TIMEDOUT" );
    }
    timeout.tv_sec += RelTimeout.tv_sec;
    timeout.tv_nsec += RelTimeout.tv_nsec;
    normalize( timeout );

    switch( Mode )
    {
    case General::ReadWriteLock::READ:
      retval = pthread_rwlock_timedrdlock( Lock, &timeout );
      break;
    case General::ReadWriteLock::WRITE:
      retval = pthread_rwlock_timedwrlock( Lock, &timeout );
      break;
    }
    return retval;
  }

#if DEAD_LOCK_DETECTOR_ENABLED
  // If a pending deadlock is detected, then an exception will be thrown.
  inline void
  log_lock( General::DeadLockDetector::state_info_type& Info )
  {
    General::DeadLockDetector::SetState( Info );
  }

  inline void
  log_lock( General::DeadLockDetector::state_info_type& Info,
	    General::DeadLockDetector::state_type State,
	    int	Error = 0 )
  {
    Info.s_state = State;
    Info.s_error = Error;

    General::DeadLockDetector::SetState( Info );
  }

#else	/* DEAD_LOCK_DETECTOR_ENABLED */

  inline void
  log_lock( General::DeadLockDetector::state_info_type& Info )
  {
  }

  inline void
  log_lock( General::DeadLockDetector::state_info_type& Info,
	    General::DeadLockDetector::state_type State,
	    int	Error = 0 )
  {
  }

#endif	/* DEAD_LOCK_DETECTOR_ENABLED */
}

namespace General
{
  //=====================================================================
  // ReadWriteLock::Sync
  //=====================================================================
  ReadWriteLock::Sync::lock_type ReadWriteLock::Sync::INITIALIZER
    = PTHREAD_RWLOCK_INITIALIZER;

  ReadWriteLock::Sync::
  Sync( )
    : m_lock( INITIALIZER ),
      m_lock_ptr( &m_lock )
  {
  }

  ReadWriteLock::Sync::
  Sync( lock_type& Source )
    : m_lock_ptr( &Source )
  {
  }

  ReadWriteLock::Sync::
  Sync( lock_type* Source )
    : m_lock_ptr( Source )
  {
  }
}

ReadWriteLock::BusyError::
BusyError( )
  : std::runtime_error( msg( ) )
{
}

std::string ReadWriteLock::BusyError::
msg( )
{
  char	buffer[ 1024];
  strerror_r( EBUSY, buffer, sizeof( buffer ) );
  return std::string( buffer );
}

ReadWriteLock::TimedoutError::
TimedoutError( )
  : std::runtime_error( msg( ) )
{
}

std::string ReadWriteLock::TimedoutError::
msg( )
{
  char	buffer[ 1024];
  strerror_r( ETIMEDOUT, buffer, sizeof( buffer ) );
  return std::string( buffer );
}

ReadWriteLock::
ReadWriteLock( lock_type& Lock, mode_type Mode,
	       const char* Filename,
	       int Linenum )
  : m_lock( &Lock ),
    m_mode( Mode )
{
  DeadLockDetector::state_info_type
    info( m_lock,
	  ( Mode == READ )
	  ? DeadLockDetector::RW_READ
	  : DeadLockDetector::RW_WRITE,
	  DeadLockDetector::PENDING,
	  Filename,
	  Linenum );

  log_lock( info );

  static const int retry_count_max = 10;

  int retry_count = 0;
  pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &m_old_cancel_state );

  try
  {
  retry:
    int retval = get_lock( m_lock, m_mode, false /* Try Lock */ );

    switch ( retval )
    {
    case 0:
      log_lock( info, DeadLockDetector::ACQUIRED );
      break;
    case EDEADLK:
      if ( retry_count < retry_count_max )
      {
	struct timespec		sleep;
	++retry_count;
	sleep.tv_sec = 0;
	sleep.tv_nsec = 50000;
	nanosleep( &sleep, &sleep );
	goto retry;	// :TRICKY: Goto for error recovery
      }
    case EBUSY:
      log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
      throw BusyError( );
      break;
    default:
      log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
      std::string msg = log( retval, "constructor:" );
      throw std::runtime_error( msg );
      break;
    }
  }
  catch( ... )
  {
    pthread_setcancelstate( m_old_cancel_state, &m_old_cancel_state );
    throw;	// rethrow the error.
  }
}

ReadWriteLock::
ReadWriteLock( lock_type& Lock, mode_type Mode, bool TryLock,
	       const char* Filename,
	       int Linenum )
  : m_lock( &Lock ),
    m_mode( Mode )
{
  DeadLockDetector::state_info_type
    info( m_lock,
	  ( Mode == READ )
	  ? DeadLockDetector::RW_READ
	  : DeadLockDetector::RW_WRITE,
	  DeadLockDetector::PENDING,
	  Filename,
	  Linenum );

  log_lock( info );

  static const int retry_count_max = 10;

  int retry_count = 0;
  pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &m_old_cancel_state );
  
  try
  {
  retry:
    int retval = get_lock( m_lock, m_mode, TryLock );

    switch ( retval )
    {
    case 0:
      log_lock( info, DeadLockDetector::ACQUIRED );
      break;
    case EDEADLK:
      if ( retry_count < retry_count_max )
      {
	log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
	struct timespec		sleep;
	++retry_count;
	sleep.tv_sec = 0;
	sleep.tv_nsec = 50000;
	nanosleep( &sleep, &sleep );
	goto retry;	// :TRICKY: Goto for error recovery
      }
    case EBUSY:
      log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
      throw BusyError( );
      break;
    default:
      log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
      std::string msg = log( retval, "constructor:" );
      throw std::runtime_error( msg );
      break;
    }
  }
  catch( ... )
  {
    pthread_setcancelstate( m_old_cancel_state, &m_old_cancel_state );
    throw;	// rethrow the error.
  }
}

ReadWriteLock::
ReadWriteLock( lock_type& Lock, mode_type Mode, int TimeoutMS,
	       const char* Filename,
	       int Linenum )
  : m_lock( &Lock ),
    m_mode( Mode )
{
  pthread_setcancelstate( PTHREAD_CANCEL_DISABLE, &m_old_cancel_state );
  DeadLockDetector::state_info_type
    info( m_lock,
	  ( Mode == READ )
	  ? DeadLockDetector::RW_READ
	  : DeadLockDetector::RW_WRITE,
	  DeadLockDetector::PENDING,
	  Filename,
	  Linenum );

  int retval = 0;

  if ( TimeoutMS <= 0 )
  {
    log_lock( info );
    retval = get_lock( m_lock, m_mode, false /* TryLock */ );
  }
  else
  {
    //-------------------------------------------------------------------
    // Non-blocking mode
    //-------------------------------------------------------------------
    struct timespec	stop;
#if VERBOSE_RETRY_LOGGING
    int			iteration = 0;
#endif /* VERBOSE_RETRY_LOGGING */

    clock_gettime( CLOCK_REALTIME, &stop );
    stop.tv_sec += TimeoutMS / 1000;
    stop.tv_nsec += ( TimeoutMS % 1000 ) * 1000000;
    normalize( stop );

#if VERBOSE_RETRY_LOGGING
    MutexLock::lock_type*	io_baton( IOLock::GetKey( std::cerr ) );
#endif /* VERBOSE_RETRY_LOGGING */

    while ( true )
    {
      try
      {
	log_lock( info, DeadLockDetector::PENDING );
	retval = get_lock( m_lock, m_mode, LockAcquisitionInterval, stop );
	if ( retval == 0 )
	{
#if VERBOSE_RETRY_LOGGING
	  if ( iteration )
	  {
	    MutexLock::Lock( io_baton );
	    std::cerr << "Succeeded in acquiring lock after " << iteration
		      << " tries"
		      << std::endl;
	    MutexLock::UnLock( io_baton );
	  }
#endif /* VERBOSE_RETRY_LOGGING */
	  break;
	}
	else
	{
	  log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
	}
	iteration++;
      }
      catch( ... )
      {
#if VERBOSE_RETRY_LOGGING
	if ( iteration )
	{
	  MutexLock::Lock( io_baton );
	  std::cerr << "Failed to acquire lock after " << iteration
		    << " tries"
		    << std::endl;
	  MutexLock::UnLock( io_baton );
	}
#endif /* VERBOSE_RETRY_LOGGING */
	retval = ETIMEDOUT;
	break;
      }
    }
  }
  try
  {
    switch ( retval )
    {
    case 0:
      log_lock( info, DeadLockDetector::ACQUIRED );
      break;
    case EBUSY:
      log( retval, "constructor(lock,mode,timeout):" );
      log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
      throw BusyError( );
      break;
    case ETIMEDOUT:
      log( retval, "constructor(lock,mode,timeout): TIMEDOUT" );
      log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
      throw TimedoutError( );
      break;
    default:
      log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
      std::string msg = log( retval, "constructor(lock,mode,timeout):" );
      throw std::runtime_error( msg );
      break;
    }
  }
  catch( ... )
  {
    pthread_setcancelstate( m_old_cancel_state, &m_old_cancel_state );
    throw;	// rethrow the error.
  }
}

ReadWriteLock::
~ReadWriteLock( )
{
  Release( __FILE__, __LINE__ );
}


void General::ReadWriteLock::
Destroy( lock_type& Lock )
{
  int retval = 0;
  retval = pthread_rwlock_destroy( &Lock );
  if ( retval != 0 )
  {
    log( retval, "destroy:" );
  }
}

void General::ReadWriteLock::
Initialize( lock_type& Lock )
{
  int retval = 0;
  retval = pthread_rwlock_init( &Lock,
				static_cast< pthread_rwlockattr_t* >( NULL ) );
  if ( retval != 0 )
  {
    log( retval, "initialize:" );
  }
}

void
ReadWriteLock::
Release( const char* Filename, int Linenum )
{
  int retval = 0;

  if ( m_lock != (lock_type*)NULL )
  {
    retval = pthread_rwlock_unlock( m_lock );

    pthread_setcancelstate( m_old_cancel_state, &m_old_cancel_state );
  }
  DeadLockDetector::state_info_type
    info( m_lock,
	  ( m_mode == READ )
	  ? DeadLockDetector::RW_READ
	  : DeadLockDetector::RW_WRITE,
	  DeadLockDetector::RELEASED,
	  __FILE__,
	  __LINE__ );
  if ( retval != 0 )
  {
    log_lock( info, DeadLockDetector::ACQUISITION_ERROR, retval );
    log( retval, "release:" );
  }
  else
  {
    log_lock( info );
    m_lock = (lock_type*)NULL;
  }
}

int ReadWriteLock::
Interval( int Value )
{
  int retval;
  retval = LockAcquisitionInterval.tv_sec * 1000;
  retval += LockAcquisitionInterval.tv_nsec / 1000000;

  LockAcquisitionInterval.tv_sec = Value / 1000;
  LockAcquisitionInterval.tv_nsec = ( Value % 1000 ) * 1000000;

  return retval;
}

